import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class CuspL(UGen):
    """
    A linear-interpolating cusp map chaotic generator.

    ::

        >>> cusp_l = supriya.ugens.CuspL.ar(
        ...     a=1,
        ...     b=1.9,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> cusp_l
        CuspL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.0), ("b", 1.9), ("xi", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class CuspN(UGen):
    """
    A non-interpolating cusp map chaotic generator.

    ::

        >>> cusp_n = supriya.ugens.CuspN.ar(
        ...     a=1,
        ...     b=1.9,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> cusp_n
        CuspN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.0), ("b", 1.9), ("xi", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class FBSineC(UGen):
    """
    A cubic-interpolating feedback sine with chaotic phase indexing.

    ::

        >>> fbsine_c = supriya.ugens.FBSineC.ar(
        ...     a=1.1,
        ...     c=0.5,
        ...     fb=0.1,
        ...     frequency=22050,
        ...     im=1,
        ...     xi=0.1,
        ...     yi=0.1,
        ... )
        >>> fbsine_c
        FBSineC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("im", 1.0),
            ("fb", 0.1),
            ("a", 1.1),
            ("c", 0.5),
            ("xi", 0.1),
            ("yi", 0.1),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class FBSineL(UGen):
    """
    A linear-interpolating feedback sine with chaotic phase indexing.

    ::

        >>> fbsine_l = supriya.ugens.FBSineL.ar(
        ...     a=1.1,
        ...     c=0.5,
        ...     fb=0.1,
        ...     frequency=22050,
        ...     im=1,
        ...     xi=0.1,
        ...     yi=0.1,
        ... )
        >>> fbsine_l
        FBSineL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("im", 1.0),
            ("fb", 0.1),
            ("a", 1.1),
            ("c", 0.5),
            ("xi", 0.1),
            ("yi", 0.1),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class FBSineN(UGen):
    """
    A non-interpolating feedback sine with chaotic phase indexing.

    ::

        >>> fbsine_n = supriya.ugens.FBSineN.ar(
        ...     a=1.1,
        ...     c=0.5,
        ...     fb=0.1,
        ...     frequency=22050,
        ...     im=1,
        ...     xi=0.1,
        ...     yi=0.1,
        ... )
        >>> fbsine_n
        FBSineN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("im", 1.0),
            ("fb", 0.1),
            ("a", 1.1),
            ("c", 0.5),
            ("xi", 0.1),
            ("yi", 0.1),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class GbmanL(UGen):
    """
    A non-interpolating gingerbreadman map chaotic generator.

    ::

        >>> gbman_l = supriya.ugens.GbmanL.ar(
        ...     frequency=22050,
        ...     xi=1.2,
        ...     yi=2.1,
        ... )
        >>> gbman_l
        GbmanL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("xi", 1.2), ("yi", 2.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class GbmanN(UGen):
    """
    A non-interpolating gingerbreadman map chaotic generator.

    ::

        >>> gbman_n = supriya.ugens.GbmanN.ar(
        ...     frequency=22050,
        ...     xi=1.2,
        ...     yi=2.1,
        ... )
        >>> gbman_n
        GbmanN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("xi", 1.2), ("yi", 2.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class HenonC(UGen):
    """
    A cubic-interpolating henon map chaotic generator.

    ::

        >>> henon_c = supriya.ugens.HenonC.ar(
        ...     a=1.4,
        ...     b=0.3,
        ...     frequency=22050,
        ...     x_0=0,
        ...     x_1=0,
        ... )
        >>> henon_c
        HenonC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.4), ("b", 0.3), ("x_0", 0), ("x_1", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class HenonL(UGen):
    """
    A linear-interpolating henon map chaotic generator.

    ::

        >>> henon_l = supriya.ugens.HenonL.ar(
        ...     a=1.4,
        ...     b=0.3,
        ...     frequency=22050,
        ...     x_0=0,
        ...     x_1=0,
        ... )
        >>> henon_l
        HenonL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.4), ("b", 0.3), ("x_0", 0), ("x_1", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class HenonN(UGen):
    """
    A non-interpolating henon map chaotic generator.

    ::

        >>> henon_n = supriya.ugens.HenonN.ar(
        ...     a=1.4,
        ...     b=0.3,
        ...     frequency=22050,
        ...     x_0=0,
        ...     x_1=0,
        ... )
        >>> henon_n
        HenonN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.4), ("b", 0.3), ("x_0", 0), ("x_1", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LatoocarfianC(UGen):
    """
    A cubic-interpolating Latoocarfian chaotic generator.

    ::

        >>> latoocarfian_c = supriya.ugens.LatoocarfianC.ar(
        ...     a=1,
        ...     b=3,
        ...     c=0.5,
        ...     d=0.5,
        ...     frequency=22050,
        ...     xi=0.5,
        ...     yi=0.5,
        ... )
        >>> latoocarfian_c
        LatoocarfianC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("a", 1),
            ("b", 3),
            ("c", 0.5),
            ("d", 0.5),
            ("xi", 0.5),
            ("yi", 0.5),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LatoocarfianL(UGen):
    """
    A linear-interpolating Latoocarfian chaotic generator.

    ::

        >>> latoocarfian_l = supriya.ugens.LatoocarfianL.ar(
        ...     a=1,
        ...     b=3,
        ...     c=0.5,
        ...     d=0.5,
        ...     frequency=22050,
        ...     xi=0.5,
        ...     yi=0.5,
        ... )
        >>> latoocarfian_l
        LatoocarfianL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("a", 1),
            ("b", 3),
            ("c", 0.5),
            ("d", 0.5),
            ("xi", 0.5),
            ("yi", 0.5),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LatoocarfianN(UGen):
    """
    A non-interpolating Latoocarfian chaotic generator.

    ::

        >>> latoocarfian_n = supriya.ugens.LatoocarfianN.ar(
        ...     a=1,
        ...     b=3,
        ...     c=0.5,
        ...     d=0.5,
        ...     frequency=22050,
        ...     xi=0.5,
        ...     yi=0.5,
        ... )
        >>> latoocarfian_n
        LatoocarfianN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("a", 1),
            ("b", 3),
            ("c", 0.5),
            ("d", 0.5),
            ("xi", 0.5),
            ("yi", 0.5),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LinCongC(UGen):
    """
    A cubic-interpolating linear congruential chaotic generator.

    ::

        >>> lin_cong_c = supriya.ugens.LinCongC.ar(
        ...     a=1.1,
        ...     c=0.13,
        ...     frequency=22050,
        ...     m=1,
        ...     xi=0,
        ... )
        >>> lin_cong_c
        LinCongC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.1), ("c", 0.13), ("m", 1), ("xi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LinCongL(UGen):
    """
    A linear-interpolating linear congruential chaotic generator.

    ::

        >>> lin_cong_l = supriya.ugens.LinCongL.ar(
        ...     a=1.1,
        ...     c=0.13,
        ...     frequency=22050,
        ...     m=1,
        ...     xi=0,
        ... )
        >>> lin_cong_l
        LinCongL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.1), ("c", 0.13), ("m", 1), ("xi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LinCongN(UGen):
    """
    A non-interpolating linear congruential chaotic generator.

    ::

        >>> lin_cong_n = supriya.ugens.LinCongN.ar(
        ...     a=1.1,
        ...     c=0.13,
        ...     frequency=22050,
        ...     m=1,
        ...     xi=0,
        ... )
        >>> lin_cong_n
        LinCongN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1.1), ("c", 0.13), ("m", 1), ("xi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LorenzL(UGen):
    """
    A linear-interpolating Lorenz chaotic generator.

    ::

        >>> lorenz_l = supriya.ugens.LorenzL.ar(
        ...     b=2.667,
        ...     frequency=22050,
        ...     h=0.05,
        ...     r=28,
        ...     s=10,
        ...     xi=0.1,
        ...     yi=0,
        ...     zi=0,
        ... )
        >>> lorenz_l
        LorenzL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 22050),
            ("s", 10),
            ("r", 28),
            ("b", 2.667),
            ("h", 0.05),
            ("xi", 0.1),
            ("yi", 0),
            ("zi", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class QuadC(UGen):
    """
    A cubic-interpolating general quadratic map chaotic generator.

    ::

        >>> quad_c = supriya.ugens.QuadC.ar(
        ...     a=1,
        ...     b=-1,
        ...     c=-0.75,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> quad_c
        QuadC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1), ("b", -1), ("c", -0.75), ("xi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class QuadL(UGen):
    """
    A linear-interpolating general quadratic map chaotic generator.

    ::

        >>> quad_l = supriya.ugens.QuadL.ar(
        ...     a=1,
        ...     b=-1,
        ...     c=-0.75,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> quad_l
        QuadL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1), ("b", -1), ("c", -0.75), ("xi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class QuadN(UGen):
    """
    A non-interpolating general quadratic map chaotic generator.

    ::

        >>> quad_n = supriya.ugens.QuadN.ar(
        ...     a=1,
        ...     b=-1,
        ...     c=-0.75,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> quad_n
        QuadN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("a", 1), ("b", -1), ("c", -0.75), ("xi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class StandardL(UGen):
    """
    A linear-interpolating standard map chaotic generator.

    ::

        >>> standard_l = supriya.ugens.StandardL.ar(
        ...     frequency=22050,
        ...     k=1,
        ...     xi=0.5,
        ...     yi=0,
        ... )
        >>> standard_l
        StandardL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("k", 1), ("xi", 0.5), ("yi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class StandardN(UGen):
    """
    A non-interpolating standard map chaotic generator.

    ::

        >>> standard_n = supriya.ugens.StandardN.ar(
        ...     frequency=22050,
        ...     k=1,
        ...     xi=0.5,
        ...     yi=0,
        ... )
        >>> standard_n
        StandardN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 22050), ("k", 1), ("xi", 0.5), ("yi", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)



import supriya.synthdefs
import supriya.ugens


def _build_synthdef():

    builder = supriya.synthdefs.SynthDefBuilder(bus=0, cutoff=0.5)
    with builder:
        source = supriya.ugens.In.ar(bus=builder["bus"])
        cutoff = builder["cutoff"]
        cutoff = supriya.ugens.Lag.kr(source=cutoff)
        lpf_cutoff = cutoff.clip(0.0, 0.75)
        lpf_frequency = lpf_cutoff.scale(0, 0.75, 20, 22000, True)
        hpf_cutoff = cutoff.clip(0.25, 1.0)
        hpf_frequency = hpf_cutoff.scale(0.25, 1.0, 20, 22000, True)
        source = supriya.ugens.LPF.ar(source=source, frequency=lpf_frequency)
        source = supriya.ugens.HPF.ar(source=source, frequency=hpf_frequency)
        supriya.ugens.ReplaceOut.ar(bus=builder["bus"], source=source)
    synthdef = builder.build()
    return synthdef


sweep_filter = _build_synthdef()

__all__ = ("sweep_filter",)



import enum
import inspect
from typing import List, Tuple, cast

from uqbar.apis import ModuleDocumenter, SummarizingClassDocumenter


class SupriyaClassDocumenter(SummarizingClassDocumenter):

    ignored_special_methods: Tuple[str, ...] = (
        "__delattr__",
        "__dict__",
        "__eq__",
        "__getattribute__",
        "__getnewargs__",
        "__getstate__",
        "__hash__",
        "__init__",
        "__new__",
        "__postinit__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__setattr__",
        "__setstate__",
        "__sizeof__",
        "__str__",
        "__subclasshook__",
        "fromkeys",
        "pipe_cloexec",
    )

    def __str__(self) -> str:
        name = getattr(self.client, "__name__")
        if issubclass(self.client, Exception):  # type: ignore
            return ".. autoexception:: {}".format(name)
        result = [".. autoclass:: {}".format(name), "   :show-inheritance:"]
        if issubclass(self.client, enum.Enum):  # type: ignore
            result.extend(["   :members:", "   :undoc-members:"])
        result.append("")
        for attr in sorted(
            inspect.classify_class_attrs(cast(type, self.client)), key=lambda x: x.name
        ):
            if attr.defining_class is not self.client:
                continue
            if attr.name.startswith("_") and not attr.name.startswith("__"):
                continue
            if attr.name.startswith("__") and attr.name in self.ignored_special_methods:
                continue
            if attr.kind in ("method", "class method", "static method"):
                result.append(f"   .. automethod:: {attr.name}")
            elif attr.kind in ("property"):
                result.append(f"   .. autoproperty:: {attr.name}")
        return "\n".join(result)


class SupriyaModuleDocumenter(ModuleDocumenter):
    def _build_toc(self, documenters, **kwargs) -> List[str]:
        result: List[str] = []
        if not documenters:
            return result
        result.extend(["", ".. toctree::", "   :hidden:"])
        result.append("")
        module_documenters = [_ for _ in documenters if isinstance(_, type(self))]
        for module_documenter in module_documenters:
            path = self._build_toc_path(module_documenter)
            if path:
                result.append("   {}".format(path))
        return result



from .Interval import Interval


class _CInterval:

    __slots__ = ("start_offset", "stop_offset", "original_interval")

    def __init__(self, start_offset, stop_offset, original_interval):
        self.start_offset = float(start_offset)
        self.stop_offset = float(stop_offset)
        self.original_interval = original_interval

    def __eq__(self, cinterval):
        assert isinstance(cinterval, type(self)), cinterval
        if self.start_offset != cinterval.start_offset:
            return False
        if self.stop_offset != cinterval.stop_offset:
            return False
        if self.original_interval != cinterval.original_interval:
            return False
        return True

    def __ne__(self, cinterval):
        assert isinstance(cinterval, type(self)), cinterval
        return not self.__eq__(cinterval)

    def __lt__(self, cinterval):
        assert isinstance(cinterval, type(self)), cinterval
        if self.start_offset < cinterval.start_offset:
            return True
        if self.start_offset > cinterval.start_offset:
            return False
        if self.stop_offset < cinterval.stop_offset:
            return True
        return False

    def __repr__(self):
        return "<_CInterval {}:{}>".format(self.start_offset, self.stop_offset)

    @classmethod
    def from_interval(cls, interval):
        start_offset = float(interval.start_offset)
        stop_offset = float(interval.stop_offset)
        return cls(start_offset, stop_offset, interval)

    def intersects_interval(self, expr):
        return (
            expr.start_offset <= self.start_offset
            and self.start_offset < expr.stop_offset
        ) or (
            self.start_offset <= expr.start_offset
            and expr.start_offset < self.stop_offset
        )


class _CNode:

    __slots__ = (
        "balance",
        "height",
        "left_child",
        "node_start_index",
        "node_stop_index",
        "payload",
        "right_child",
        "start_offset",
        "stop_offset_high",
        "stop_offset_low",
        "subtree_start_index",
        "subtree_stop_index",
    )

    def __init__(self, start_offset):
        self.balance = 0
        self.height = 0
        self.left_child = None
        self.node_start_index = -1
        self.node_stop_index = -1
        self.payload = []
        self.right_child = None
        self.start_offset = start_offset
        self.stop_offset_high = None
        self.stop_offset_low = None
        self.subtree_start_index = -1
        self.subtree_stop_index = -1


class IntervalTreeDriver:

    ### CLASS VARIABLES ###

    __slots__ = ("_root_node",)

    ### INITIALIZER ###

    def __init__(self, intervals=None):
        self._root_node = None
        self.update(intervals or [])

    ### SPECIAL METHODS ###

    def __contains__(self, interval):
        if not self._is_interval(interval):
            raise ValueError(interval)
        candidates = self.find_intervals_starting_at(interval.start_offset)
        result = interval in candidates
        return result

    def __getitem__(self, item):
        if isinstance(item, int):
            if self._root_node is None:
                raise IndexError
            if item < 0:
                item = self._root_node.subtree_stop_index + item
            if item < 0 or self._root_node.subtree_stop_index <= item:
                raise IndexError
            cinterval = self._recurse_getitem_by_index(self._root_node, item)
            return cinterval.original_interval
        elif isinstance(item, slice):
            if self._root_node is None:
                return []
            indices = item.indices(self._root_node.subtree_stop_index)
            start, stop = indices[0], indices[1]
            cintervals = self._recurse_getitem_by_slice(self._root_node, start, stop)
            return [cinterval.original_interval for cinterval in cintervals]
        raise TypeError("Indices must be integers or slices, got {}".format(item))

    def __iter__(self):
        stack = []
        current = self._root_node
        while True:
            while current is not None:
                stack.append(current)
                current = current.left_child
            if not stack:
                return
            current = stack.pop()
            for cinterval in current.payload:
                yield cinterval.original_interval
            while current.right_child is None and stack:
                current = stack.pop()
                for cinterval in current.payload:
                    yield cinterval.original_interval
            current = current.right_child

    def __len__(self):
        if self._root_node is None:
            return 0
        return self._root_node.subtree_stop_index

    ### PRIVATE METHODS ###

    def _get_node_cinterval(self, node):
        return Interval(
            start_offset=node.start_offset, stop_offset=node.stop_offset_high
        )

    def _insert_node(self, node, start_offset):
        if node is None:
            return _CNode(start_offset)
        if start_offset < node.start_offset:
            left_child = self._insert_node(node.left_child, start_offset)
            self._set_node_left_child(node, left_child)
        elif node.start_offset < start_offset:
            right_child = self._insert_node(node.right_child, start_offset)
            self._set_node_right_child(node, right_child)
        return self._rebalance(node)

    def _insert_interval(self, cinterval):
        self._root_node = self._insert_node(self._root_node, cinterval.start_offset)
        node = self._search(self._root_node, cinterval.start_offset)
        node.payload.append(cinterval)
        node.payload.sort(key=lambda x: x.stop_offset)

    @staticmethod
    def _is_interval(expr):
        if hasattr(expr, "start_offset") and hasattr(expr, "stop_offset"):
            return True
        return False

    def _remove_node(self, node, start_offset):
        if node is None:
            return None
        if node.start_offset == start_offset:
            if node.left_child and node.right_child:
                next_node = node.right_child
                while next_node.left_child:
                    next_node = next_node.left_child
                node.start_offset = next_node.start_offset
                node.payload = next_node.payload
                self._set_node_right_child(
                    node, self._remove_node(node.right_child, next_node.start_offset)
                )
            else:
                node = node.left_child or node.right_child
        elif start_offset < node.start_offset:
            left_child = self._remove_node(node.left_child, start_offset)
            self._set_node_left_child(node, left_child)
        elif node.start_offset < start_offset:
            self._set_node_right_child(
                node, self._remove_node(node.right_child, start_offset)
            )
        return self._rebalance(node)

    def _rebalance(self, node):
        if node is None:
            return None
        if 1 < node.balance:
            if 0 <= node.right_child.balance:
                node = self._rotate_right_right(node)
            else:
                node = self._rotate_right_left(node)
        elif node.balance < -1:
            if node.left_child.balance <= 0:
                node = self._rotate_left_left(node)
            else:
                node = self._rotate_left_right(node)
        return node

    def _recurse_find_intervals_intersecting_interval(self, node, cinterval):
        result = []
        if node is None:
            return result
        node_interval = self._get_node_cinterval(node)
        if cinterval.intersects_interval(node_interval):
            subresult = self._recurse_find_intervals_intersecting_interval(
                node.left_child, cinterval
            )
            result.extend(subresult)
            for candidate_interval in node.payload:
                if candidate_interval.intersects_interval(cinterval):
                    result.append(candidate_interval)
            subresult = self._recurse_find_intervals_intersecting_interval(
                node.right_child, cinterval
            )
            result.extend(subresult)
        elif (cinterval.start_offset <= node.start_offset) or (
            cinterval.stop_offset <= node.start_offset
        ):
            subresult = self._recurse_find_intervals_intersecting_interval(
                node.left_child, cinterval
            )
            result.extend(subresult)
        return result

    def _recurse_find_intervals_intersecting_offset(self, node, offset):
        result = []
        if node is None:
            return result
        if node.start_offset <= offset < node.stop_offset_high:
            subresult = self._recurse_find_intervals_intersecting_offset(
                node.left_child, offset
            )
            result.extend(subresult)
            for cinterval in node.payload:
                if offset < cinterval.stop_offset:
                    result.append(cinterval)
            subresult = self._recurse_find_intervals_intersecting_offset(
                node.right_child, offset
            )
            result.extend(subresult)
        elif offset <= node.start_offset:
            subresult = self._recurse_find_intervals_intersecting_offset(
                node.left_child, offset
            )
            result.extend(subresult)
        return result

    def _recurse_find_intervals_stopping_at(self, node, offset):
        result = []
        if node is None:
            return result
        if node.stop_offset_low <= offset <= node.stop_offset_high:
            for cinterval in node.payload:
                if cinterval.stop_offset == offset:
                    result.append(cinterval)
            if node.left_child is not None:
                result.extend(
                    self._recurse_find_intervals_stopping_at(node.left_child, offset)
                )
            if node.right_child is not None:
                result.extend(
                    self._recurse_find_intervals_stopping_at(node.right_child, offset)
                )
        return result

    def _recurse_get_start_offset_after(self, node, offset):
        if node is None:
            return None
        result = None
        if node.start_offset <= offset and node.right_child:
            result = self._recurse_get_start_offset_after(node.right_child, offset)
        elif offset < node.start_offset:
            result = (
                self._recurse_get_start_offset_after(node.left_child, offset) or node
            )
        return result

    def _recurse_get_start_offset_before(self, node, offset):
        if node is None:
            return None
        result = None
        if node.start_offset < offset:
            result = (
                self._recurse_get_start_offset_before(node.right_child, offset) or node
            )
        elif offset <= node.start_offset and node.left_child:
            result = self._recurse_get_start_offset_before(node.left_child, offset)
        return result

    def _recurse_getitem_by_index(self, node, index):
        if node.node_start_index <= index < node.node_stop_index:
            return node.payload[index - node.node_start_index]
        elif node.left_child is not None and index < node.node_start_index:
            return self._recurse_getitem_by_index(node.left_child, index)
        elif node.right_child is not None and node.node_stop_index <= index:
            return self._recurse_getitem_by_index(node.right_child, index)

    def _recurse_getitem_by_slice(self, node, start, stop):
        result = []
        if node is None:
            return result
        if start < node.node_start_index and node.left_child is not None:
            result.extend(self._recurse_getitem_by_slice(node.left_child, start, stop))
        if start < node.node_stop_index and node.node_start_index < stop:
            node_start = start - node.node_start_index
            if node_start < 0:
                node_start = 0
            node_stop = stop - node.node_start_index
            result.extend(node.payload[node_start:node_stop])
        if node.node_stop_index <= stop and node.right_child is not None:
            result.extend(self._recurse_getitem_by_slice(node.right_child, start, stop))
        return result

    def _remove_interval(self, cinterval, old_start_offset=None):
        assert isinstance(cinterval, _CInterval)
        start_offset = cinterval.start_offset
        if old_start_offset is not None:
            start_offset = old_start_offset
        node = self._search(self._root_node, start_offset)
        if node is None:
            return
        if cinterval in node.payload:
            node.payload.remove(cinterval)
        if not node.payload:
            self._root_node = self._remove_node(self._root_node, start_offset)

    def _rotate_left_left(self, node):
        next_node = node.left_child
        self._set_node_left_child(node, next_node.right_child)
        self._set_node_right_child(next_node, node)
        return next_node

    def _rotate_left_right(self, node):
        self._set_node_left_child(node, self._rotate_right_right(node.left_child))
        next_node = self._rotate_left_left(node)
        return next_node

    def _rotate_right_left(self, node):
        self._set_node_right_child(node, self._rotate_left_left(node.right_child))
        next_node = self._rotate_right_right(node)
        return next_node

    def _rotate_right_right(self, node):
        next_node = node.right_child
        self._set_node_right_child(node, next_node.left_child)
        self._set_node_left_child(next_node, node)
        return next_node

    def _search(self, node, start_offset):
        if node is None:
            return None
        if node.start_offset == start_offset:
            return node
        elif node.left_child and start_offset < node.start_offset:
            return self._search(node.left_child, start_offset)
        elif node.right_child and node.start_offset < start_offset:
            return self._search(node.right_child, start_offset)
        return None

    def _set_node_left_child(self, node, left_child):
        node.left_child = left_child
        self._update_node_height(node)

    def _set_node_right_child(self, node, right_child):
        node.right_child = right_child
        self._update_node_height(node)

    def _update_node_height(self, node):
        left_height = -1
        right_height = -1
        if node.left_child is not None:
            left_height = node.left_child.height
        if node.right_child is not None:
            right_height = node.right_child.height
        node.height = max(left_height, right_height) + 1
        node.balance = right_height - left_height
        return node.height

    def _update_indices(self, node, parent_stop_index):
        if node is None:
            return
        if node.left_child is not None:
            self._update_indices(node.left_child, parent_stop_index)
            node.node_start_index = node.left_child.subtree_stop_index
            node.subtree_start_index = node.left_child.subtree_start_index
        elif parent_stop_index == -1:
            node.node_start_index = 0
            node.subtree_start_index = 0
        else:
            node.node_start_index = parent_stop_index
            node.subtree_start_index = parent_stop_index
        node.node_stop_index = node.node_start_index + len(node.payload)
        node.subtree_stop_index = node.node_stop_index
        if node.right_child is not None:
            self._update_indices(node.right_child, node.node_stop_index)
            node.subtree_stop_index = node.right_child.subtree_stop_index

    def _update_offsets(self, node):
        if node is None:
            return
        stop_offset_low = min(x.stop_offset for x in node.payload)
        stop_offset_high = max(x.stop_offset for x in node.payload)
        if node.left_child:
            left_child = self._update_offsets(node.left_child)
            if left_child.stop_offset_low < stop_offset_low:
                stop_offset_low = left_child.stop_offset_low
            if stop_offset_high < left_child.stop_offset_high:
                stop_offset_high = left_child.stop_offset_high
        if node.right_child:
            right_child = self._update_offsets(node.right_child)
            if right_child.stop_offset_low < stop_offset_low:
                stop_offset_low = right_child.stop_offset_low
            if stop_offset_high < right_child.stop_offset_high:
                stop_offset_high = right_child.stop_offset_high
        node.stop_offset_low = stop_offset_low
        node.stop_offset_high = stop_offset_high
        return node

    ### PUBLIC METHODS ###

    def add(self, interval):
        if not self._is_interval(interval):
            # raise ValueError(interval)
            return
        cinterval = _CInterval.from_interval(interval)
        self._insert_interval(cinterval)
        self._update_indices(self._root_node, -1)
        self._update_offsets(self._root_node)

    def find_intervals_intersecting_offset(self, offset):
        offset = float(offset)
        cintervals = self._recurse_find_intervals_intersecting_offset(
            self._root_node, offset
        )
        cintervals.sort(key=lambda x: (x.start_offset, x.stop_offset))
        return [cinterval.original_interval for cinterval in cintervals]

    def find_intervals_intersecting_interval(self, interval):
        cinterval = _CInterval.from_interval(interval)
        cintervals = self._recurse_find_intervals_intersecting_interval(
            self._root_node, cinterval
        )
        cintervals.sort(key=lambda x: (x.start_offset, x.stop_offset))
        return [cinterval.original_interval for cinterval in cintervals]

    def find_intervals_starting_at(self, offset):
        results = []
        node = self._search(self._root_node, offset)
        if node is None:
            return results
        results.extend(cinterval.original_interval for cinterval in node.payload)
        return results

    def find_intervals_stopping_at(self, offset):
        cintervals = self._recurse_find_intervals_stopping_at(self._root_node, offset)
        cintervals.sort(key=lambda x: (x.start_offset, x.stop_offset))
        return [cinterval.original_interval for cinterval in cintervals]

    def get_start_offset_after(self, offset):
        node = self._recurse_get_start_offset_after(self._root_node, offset)
        if node is None:
            return None
        return node.start_offset

    def get_start_offset_before(self, offset):
        node = self._recurse_get_start_offset_before(self._root_node, offset)
        if node is None:
            return None
        return node.start_offset

    def index(self, interval):
        assert self._is_interval(interval)
        cinterval = _CInterval.from_interval(interval)
        node = self._search(self._root_node, cinterval.start_offset)
        if node is None:
            raise ValueError("{} not in interval tree.".format(interval))
        if cinterval not in node.payload:
            raise ValueError("{} not in interval tree.".format(interval))
        index = node.payload.index(cinterval) + node.node_start_index
        return index

    def remove(self, interval):
        if interval not in self:
            raise ValueError(interval)
        cinterval = _CInterval.from_interval(interval)
        self._remove_interval(cinterval)
        self._update_indices(self._root_node, -1)
        self._update_offsets(self._root_node)

    def update(self, intervals):
        # for interval in intervals:
        #     if not self._is_interval(interval):
        #         raise ValueError(interval)
        for interval in intervals:
            if not self._is_interval(interval):
                # raise ValueError(interval)
                continue
            cinterval = _CInterval.from_interval(interval)
            self._insert_interval(cinterval)
        self._update_indices(self._root_node, -1)
        self._update_offsets(self._root_node)



import asyncio
import logging
import queue
import traceback
from typing import Optional, Tuple

from .bases import BaseClock
from .ephemera import ClockContext, EventType, Moment

logger = logging.getLogger("supriya.clocks")


class AsyncClock(BaseClock):
    def __init__(self):
        BaseClock.__init__(self)
        self._event = asyncio.Event()
        self._task = None
        self._slop = 1.0

    ### SCHEDULING METHODS ###

    def _enqueue_command(self, command):
        super()._enqueue_command(command)
        self._event.set()

    async def _perform_callback_event(self, event, current_moment, desired_moment):
        logger.debug(
            f"[{self.name}] ... ... Performing {event.procedure} at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s / "
            f"{desired_moment.offset}:o"
        )
        context = ClockContext(current_moment, desired_moment, event)
        args = event.args or ()
        kwargs = event.kwargs or {}
        try:
            result = event.procedure(context, *args, **kwargs)
            if asyncio.iscoroutine(result):
                result = await result
        except Exception:
            traceback.print_exc()
            return
        self._process_callback_event_result(desired_moment, event, result)

    async def _perform_events(self, current_moment: Moment):
        logger.debug(
            f"[{self.name}] ... Ready to perform at "
            f"{current_moment.seconds - self._state.initial_seconds}:s / "
            f"{current_moment.offset}:o"
        )
        while self._is_running and self._event_queue.qsize():
            (
                event,
                desired_moment,
                should_continue,
                should_break,
            ) = self._process_perform_event_loop(current_moment)
            if should_continue:
                continue
            elif should_break:
                break
            if event.event_type == EventType.CHANGE:
                current_moment, should_continue = self._perform_change_event(
                    event, current_moment, desired_moment
                )
                if not should_continue:
                    break
            else:
                await self._perform_callback_event(
                    event, current_moment, desired_moment
                )
                self._process_command_deque()
        return current_moment

    async def _run(self, *args, offline=False, **kwargs):
        logger.debug(f"[{self.name}] Coroutine start")
        self._process_command_deque(first_run=True)
        while self._is_running:
            logger.debug(f"[{self.name}] Loop start")
            if not await self._wait_for_queue():
                return
            try:
                current_moment = await self._wait_for_moment()
            except queue.Empty:
                continue
            if current_moment is None:
                return
            current_moment = await self._perform_events(current_moment)
            self._state = self._state._replace(
                previous_seconds=current_moment.seconds,
                previous_offset=current_moment.offset,
            )
        logger.debug(f"[{self.name}] Coroutine terminating")
        self._stop()

    async def _wait_for_event(self, sleep_time):
        try:
            await asyncio.wait_for(self._event.wait(), sleep_time)
        except (asyncio.TimeoutError, RuntimeError):
            pass

    async def _wait_for_moment(self, offline=False) -> Optional[Moment]:
        current_time = self.get_current_time()
        next_time = self._event_queue.peek().seconds
        logger.debug(
            f"[{self.name}] ... Waiting for next moment at {next_time} from {current_time}"
        )
        while current_time < next_time:
            if not offline:
                await self._wait_for_event(next_time - current_time)
            if not self._is_running:
                return None
            self._process_command_deque()
            next_time = self._event_queue.peek().seconds
            current_time = self.get_current_time()
            self._event.clear()
        return self._seconds_to_moment(current_time)

    async def _wait_for_queue(self, offline=False) -> bool:
        logger.debug(f"[{self.name}] ... Waiting for events")
        self._process_command_deque()
        self._event.clear()
        while not self._event_queue.qsize():
            if not offline:
                await self._event.wait()
            if not self._is_running:
                return False
            self._process_command_deque()
            self._event.clear()
        return True

    ### PUBLIC METHODS ###

    def cancel(self, event_id) -> Optional[Tuple]:
        event = super().cancel(event_id)
        self._event.set()
        return event

    async def start(
        self,
        initial_time: Optional[float] = None,
        initial_offset: float = 0.0,
        initial_measure: int = 1,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ):
        self._start(
            initial_time=initial_time,
            initial_offset=initial_offset,
            initial_measure=initial_measure,
            beats_per_minute=beats_per_minute,
            time_signature=time_signature,
        )
        loop = asyncio.get_running_loop()
        self._task = loop.create_task(self._run())

    async def stop(self):
        if self._stop():
            self._event.set()
            await self._task



import collections
from collections.abc import Sequence

from supriya import CalculationRate, utils
from supriya.synthdefs import PureUGen, UGen


class Blip(UGen):
    """
    A band limited impulse generator.

    ::

        >>> blip = supriya.ugens.Blip.ar(
        ...     frequency=440,
        ...     harmonic_count=200,
        ... )
        >>> blip
        Blip.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("harmonic_count", 200.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class FSinOsc(UGen):
    """
    Very fast sine wave generator (2 PowerPC instructions per output sample!)
    implemented using a ringing filter.

    ::

        >>> fsin_osc = supriya.ugens.FSinOsc.ar(
        ...     frequency=440,
        ...     initial_phase=0,
        ... )
        >>> fsin_osc
        FSinOsc.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Klank(UGen):
    """
    A bank of resonators.

    ::

        >>> frequencies = [200, 671, 1153, 1723]
        >>> amplitudes = None
        >>> decay_times = [1, 1, 1, 1]
        >>> specifications = [frequencies, amplitudes, decay_times]
        >>> source = supriya.ugens.BrownNoise.ar() * 0.001
        >>> klank = supriya.ugens.Klank.ar(
        ...     decay_scale=1,
        ...     frequency_offset=0,
        ...     frequency_scale=1,
        ...     source=source,
        ...     specifications=specifications,
        ... )
        >>> klank
        Klank.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("frequency_scale", 1),
            ("frequency_offset", 0),
            ("decay_scale", 1),
            ("specifications", None),
        ]
    )
    _unexpanded_input_names = ("specifications",)
    _valid_calculation_rates = (CalculationRate.AUDIO,)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        decay_scale=1,
        frequency_offset=0,
        frequency_scale=1,
        source=None,
        specifications=None,
    ):
        frequencies, amplitudes, decay_times = specifications
        assert len(frequencies)
        if not amplitudes:
            amplitudes = [1.0] * len(frequencies)
        elif not isinstance(amplitudes, Sequence):
            amplitudes = [amplitudes] * len(frequencies)
        if not decay_times:
            decay_times = [1.0] * len(frequencies)
        elif not isinstance(decay_times, Sequence):
            decay_times = [decay_times] * len(frequencies)
        specifications = utils.zip_sequences(frequencies, amplitudes, decay_times)
        specifications = utils.flatten_iterable(specifications)
        specifications = tuple(specifications)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            decay_scale=decay_scale,
            frequency_offset=frequency_offset,
            frequency_scale=frequency_scale,
            source=source,
            specifications=specifications,
        )


class Pulse(UGen):
    """
    Band limited pulse wave generator with pulse width modulation.

    ::

        >>> pulse = supriya.ugens.Pulse.ar(
        ...     frequency=440,
        ...     width=0.5,
        ... )
        >>> pulse
        Pulse.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 440), ("width", 0.5)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Saw(PureUGen):
    """
    A band-limited sawtooth oscillator unit generator.

    ::

        >>> supriya.ugens.Saw.ar()
        Saw.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 440.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)
